import { Component, OnDestroy, OnInit } from "@angular/core";
import { FormBuilder, FormControl, FormGroup } from "@angular/forms";
import { AuthService } from "../../../../services/auth/auth.service";
import { SpinnerService } from "../../../../services/spinner/spinner.service";
import { ToastService } from "../../../../services/toast/toast.service";
import { ConfirmationDialogService } from "../../../../services/confirmation-dialog/confirmation-dialog.service";
import { TranslateService } from "@ngx-translate/core";
import { catchError, debounceTime, takeUntil } from "rxjs/operators";
import { LazyLoadEvent, MenuItem } from "primeng/api";
import { Table } from "primeng/table";
import { User, UserFilter } from "../../../../models/data/user.model";
import { Queries } from "../../../../models/graphql/queries.model";
import { LevelListQuery, Query, UserListQuery } from "../../../../models/graphql/query.model";
import { DataService } from "../../../../services/data/data.service";
import { Page } from "../../../../models/graphql/filter/page.model";
import { Sort, SortDirection } from "../../../../models/graphql/filter/sort.model";
import { Mutations } from "../../../../models/graphql/mutations.model";
import { Subject, throwError } from "rxjs";
import { UpdateUserMutationResult } from "../../../../models/graphql/result.model";
import { ActivatedRoute } from "@angular/router";
import { Level } from "../../../../models/data/level.model";
import { Server } from "../../../../models/data/server.model";
import { ComponentWithTable } from "../../../../base/component-with-table";

@Component({
  selector: "app-members",
  templateUrl: "./members.component.html",
  styleUrls: ["./members.component.scss"]
})
export class MembersComponent extends ComponentWithTable implements OnInit, OnDestroy {
  members!: User[];
  levels!: MenuItem[];
  leftServerOptions: { label: string; value: boolean; }[] = [];
  loading = true;

  clonedUsers: { [s: string]: User; } = {};
  isEditingNew: boolean = false;

  newUserTemplate: User = {
    discordId: 0,
    name: "",
    xp: 0,
    ontime: 0,
    level: undefined,
    server: undefined,

    joinedServerCount: 0,
    joinedServers: [],

    joinedVoiceChannelCount: 0,
    joinedVoiceChannels: [],

    userJoinedGameServerCount: 0,
    userJoinedGameServers: [],

    createdAt: "",
    modifiedAt: ""
  };

  filterForm!: FormGroup<{
    id: FormControl<number | null>,
    discordId: FormControl<number | null>,
    name: FormControl<string | null>,
    leftServer: FormControl<boolean | null>,
    level: FormControl<Level | null>
  }>;

  filter: UserFilter = {};
  page: Page = {
    pageSize: undefined,
    pageIndex: undefined
  };
  sort: Sort = {
    sortColumn: undefined,
    sortDirection: undefined
  };

  totalRecords!: number;
  private unsubscriber = new Subject<void>();
  private server: Server = {};
  query = Queries.userQueryWithHistory;

  constructor(
    private authService: AuthService,
    private spinner: SpinnerService,
    private toastService: ToastService,
    private confirmDialog: ConfirmationDialogService,
    private fb: FormBuilder,
    private translate: TranslateService,
    private data: DataService,
    private route: ActivatedRoute
  ) {
    super("member", ["id", "discord_id", "name", "xp", "ontime", "left_server", "level"]);
  }

  ngOnInit(): void {
    this.setFilterForm();

    this.data.getServerFromRoute(this.route).then(server => {
      this.server = server;
      this.spinner.showSpinner();
      this.data.query<LevelListQuery>(Queries.levelQuery, {
          serverId: server.id
        },
        (data: Query) => {
          return data.servers[0];
        }
      ).subscribe(data => {
        this.levels = data.levels.map(level => {
          return { label: level.name, value: level };
        });
      });
      this.leftServerOptions = [
        { label: this.translate.instant("common.bool_as_string.true"), value: false },
        { label: this.translate.instant("common.bool_as_string.false"), value: true }
      ];
      this.loadNextPage();
    });
  }

  public ngOnDestroy(): void {
    this.unsubscriber.next();
    this.unsubscriber.complete();
  }

  loadNextPage() {
    this.spinner.showSpinner();
    this.loading = true;
    this.data.query<UserListQuery>(Queries.usersQuery, {
        serverId: this.server.id, filter: this.filter, page: this.page, sort: this.sort
      },
      (x: { servers: Server[] }) => {
        return x.servers[0];
      }
    ).subscribe(data => {
      this.totalRecords = data.userCount;
      this.members = data.users;
      this.spinner.hideSpinner();
      this.loading = false;
    });
  }

  setFilterForm() {
    this.filterForm = this.fb.group({
      id: new FormControl<number | null>(null),
      discordId: new FormControl<number | null>(null),
      name: [""],
      leftServer: new FormControl<boolean | null>(null),
      level: new FormControl<Level | null>(null)
    });

    this.filterForm.valueChanges.pipe(
      takeUntil(this.unsubscriber),
      debounceTime(600)
    ).subscribe(changes => {
      if (changes.id) {
        this.filter.id = changes.id;
      } else {
        this.filter.id = undefined;
      }

      if (changes.discordId) {
        this.filter.discordId = changes.discordId;
      } else {
        this.filter.discordId = undefined;
      }

      if (changes.name) {
        this.filter.name = changes.name;
      } else {
        this.filter.name = undefined;
      }

      if (changes.leftServer !== undefined && changes.leftServer !== null) {
        this.filter.leftServer = changes.leftServer;
      } else {
        this.filter.leftServer = undefined;
      }

      if (changes.level) {
        this.filter.level = {
          id: changes.level.id
        };
      } else {
        this.filter.level = undefined;
      }

      if (this.page.pageSize)
        this.page.pageSize = 10;

      if (this.page.pageIndex)
        this.page.pageIndex = 0;

      this.loadNextPage();
    });
  }

  nextPage(event: LazyLoadEvent) {
    this.page.pageSize = event.rows ?? 0;
    if (event.first != null && event.rows != null)
      this.page.pageIndex = event.first / event.rows;
    this.sort.sortColumn = event.sortField ?? undefined;
    this.sort.sortDirection = event.sortOrder === 1 ? SortDirection.ASC : event.sortOrder === -1 ? SortDirection.DESC : SortDirection.ASC;

    this.loadNextPage();
  }

  resetFilters() {
    this.filterForm.reset();
  }

  onRowEditInit(table: Table, user: User, index: number) {
    this.clonedUsers[index] = { ...user };
  }

  onRowEditSave(table: Table, newUser: User, index: number) {
    // const oldUser = this.clonedUsers[index];
    // delete this.clonedUsers[index];

    // if (JSON.stringify(oldUser) === JSON.stringify(newUser) && !this.isEditingNew) {
    //   console.log(1, oldUser, newUser, JSON.stringify(oldUser) === JSON.stringify(newUser), !this.isEditingNew);
    //   return;
    // }

    if (this.isEditingNew && JSON.stringify(newUser) === JSON.stringify(this.newUserTemplate)) {
      this.isEditingNew = false;
      this.members.splice(index, 1);
      return;
    }

    if (!newUser.id || !newUser.xp && !newUser.level?.id) {
      return;
    }

    this.spinner.showSpinner();
    this.data.mutation<UpdateUserMutationResult>(Mutations.updateUser, {
        id: newUser.id,
        xp: newUser.xp,
        levelId: newUser.level?.id
      }
    ).pipe(catchError(err => {
      this.spinner.hideSpinner();
      return throwError(err);
    })).subscribe(_ => {
      this.spinner.hideSpinner();
      this.toastService.success(this.translate.instant("view.server.members.message.user_changed"), this.translate.instant("view.server.members.message.user_changed_d", { name: newUser.name }));
      this.loadNextPage();
    });

  }

  onRowEditCancel(index: number) {
    if (this.isEditingNew) {
      this.members.splice(index, 1);
      delete this.clonedUsers[index];
      this.isEditingNew = false;
      return;
    }

    this.members[index] = this.clonedUsers[index];
    delete this.clonedUsers[index];
  }
}
